# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test GatesInBasis pass."""

from qiskit.circuit import QuantumCircuit, ForLoopOp, IfElseOp, SwitchCaseOp, Clbit
from qiskit.circuit.classical import expr, types
from qiskit.circuit.library import HGate, CXGate, UGate, XGate, ZGate
from qiskit.circuit.measure import Measure
from qiskit.circuit.equivalence_library import SessionEquivalenceLibrary
from qiskit.passmanager.flow_controllers import ConditionalController
from qiskit.transpiler import PassManager
from qiskit.transpiler.passes import BasisTranslator
from qiskit.transpiler.passes import GatesInBasis
from qiskit.transpiler.target import Target
from qiskit.providers.fake_provider import GenericBackendV2
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestGatesInBasisPass(QiskitTestCase):
    """Tests for GatesInBasis pass."""

    def test_all_gates_in_basis(self):
        """Test circuit with all gates in basis."""
        basis_gates = ["cx", "h"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertTrue(property_set["all_gates_in_basis"])

    def test_all_gates_not_in_basis(self):
        """Test circuit with not all gates in basis."""
        basis_gates = ["cx", "u"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_empty_circuit(self):
        """Test circuit with no gates."""
        basis_gates = ["cx", "u"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates)
        circuit = QuantumCircuit(2)
        analysis_pass(circuit, property_set=property_set)
        self.assertTrue(property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_empty_basis(self):
        """Test circuit with gates and empty basis."""
        basis_gates = []
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_after_translation(self):
        """Test circuit with gates in basis after conditional translation."""
        basis_gates = ["cx", "u"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])
        pm = PassManager()
        pm.append(analysis_pass)
        pm.append(
            ConditionalController(
                BasisTranslator(SessionEquivalenceLibrary, basis_gates),
                condition=lambda property_set: not property_set["all_gates_in_basis"],
            )
        )
        pm.append(analysis_pass)
        pm.run(circuit)
        self.assertTrue(pm.property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_with_target(self):
        """Test circuit with all gates in basis with target."""
        target = GenericBackendV2(num_qubits=5, basis_gates=["u", "cx"], seed=42).target
        basis_gates = ["cx", "u"]  # not used
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates, target=target)
        circuit = QuantumCircuit(2)
        circuit.u(0, 0, 0, 0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertTrue(property_set["all_gates_in_basis"])

    def test_all_gates_not_in_basis_with_target(self):
        """Test circuit with not all gates in basis with target."""
        target = GenericBackendV2(num_qubits=5, basis_gates=["u", "cx"], seed=42).target
        basis_gates = ["cx", "h"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates, target=target)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_not_on_all_qubits_with_target(self):
        """Test circuit with gate in global basis but not local basis."""
        target = GenericBackendV2(num_qubits=5, basis_gates=["u", "cx"], seed=42).target
        basis_gates = ["ecr", "cx", "h"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates, target=target)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.ecr(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_empty_circuit_with_target(self):
        """Test circuit with no gates with target."""
        target = GenericBackendV2(num_qubits=5, basis_gates=["u", "cx"], seed=42).target
        basis_gates = ["cx", "u"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates, target=target)
        circuit = QuantumCircuit(2)
        analysis_pass(circuit, property_set=property_set)
        self.assertTrue(property_set["all_gates_in_basis"])

    def test_all_gates_in_empty_target(self):
        """Test circuit with gates and empty basis with target."""
        target = Target()
        basis_gates = []
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates, target=target)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])

    def test_all_gates_in_ideal_sim_target(self):
        """Test with target that has ideal gates."""
        target = Target(num_qubits=2)
        target.add_instruction(HGate())
        target.add_instruction(CXGate())
        target.add_instruction(Measure())
        property_set = {}
        analysis_pass = GatesInBasis(target=target)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertTrue(property_set["all_gates_in_basis"])

    def test_all_gates_not_in_ideal_sim_target(self):
        """Test with target that has ideal gates."""
        target = Target()
        target.add_instruction(HGate())
        target.add_instruction(UGate(0, 0, 0))
        target.add_instruction(Measure())
        property_set = {}
        analysis_pass = GatesInBasis(target=target)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])

    def test_all_gates_in_basis_after_translation_with_target(self):
        """Test circuit with gates in basis after conditional translation."""
        target = GenericBackendV2(num_qubits=5, basis_gates=["u", "cx"], seed=42).target
        basis_gates = ["cx", "u"]
        property_set = {}
        analysis_pass = GatesInBasis(basis_gates, target)
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.measure_all()
        analysis_pass(circuit, property_set=property_set)
        self.assertFalse(property_set["all_gates_in_basis"])
        pm = PassManager()
        pm.append(analysis_pass)
        pm.append(
            ConditionalController(
                BasisTranslator(SessionEquivalenceLibrary, basis_gates, target=target),
                condition=lambda property_set: not property_set["all_gates_in_basis"],
            )
        )
        pm.append(analysis_pass)
        pm.run(circuit)
        self.assertTrue(pm.property_set["all_gates_in_basis"])

    def test_basis_gates_control_flow(self):
        """Test that the pass recurses into control flow."""
        circuit = QuantumCircuit(4, 1)
        circuit.h(0)
        circuit.measure(0, 0)
        with circuit.for_loop((1, 2)):
            circuit.cx(0, 1)
            with circuit.if_test((circuit.clbits[0], True)) as else_:
                circuit.x(2)
            with else_:
                circuit.z(3)

        one_missing = {"h", "measure", "for_loop", "cx", "if_else", "x"}
        pass_ = GatesInBasis(one_missing)
        pass_(circuit)
        self.assertFalse(pass_.property_set["all_gates_in_basis"])

        complete = one_missing | {"z"}
        pass_ = GatesInBasis(complete)
        pass_(circuit)
        self.assertTrue(pass_.property_set["all_gates_in_basis"])

    def test_basis_gates_target(self):
        """Test that the pass recurses into control flow."""
        circuit = QuantumCircuit(4, 1)
        circuit.h(0)
        circuit.measure(0, 0)
        with circuit.for_loop((1, 2)):
            circuit.cx(0, 1)
            with circuit.if_test((circuit.clbits[0], True)) as else_:
                circuit.x(2)
            with else_:
                circuit.z(3)

        instructions = [
            HGate(),
            Measure(),
            ForLoopOp((), None, QuantumCircuit(4)),
            CXGate(),
            IfElseOp((Clbit(), True), QuantumCircuit(2), QuantumCircuit(2)),
            SwitchCaseOp(Clbit(), [(False, QuantumCircuit(2)), (True, QuantumCircuit(2))]),
            XGate(),
            ZGate(),
        ]
        one_missing = Target(num_qubits=4)
        for instruction in instructions[:-1]:
            one_missing.add_instruction(instruction, {None: None})
        pass_ = GatesInBasis(target=one_missing)
        pass_(circuit)
        self.assertFalse(pass_.property_set["all_gates_in_basis"])

        complete = Target(num_qubits=4)
        for instruction in instructions:
            complete.add_instruction(instruction, {None: None})
        pass_ = GatesInBasis(target=complete)
        pass_(circuit)
        self.assertTrue(pass_.property_set["all_gates_in_basis"])

    def test_store_is_treated_as_builtin_basis_gates(self):
        """Test that `Store` is treated as an automatic built-in when given basis gates."""
        pass_ = GatesInBasis(basis_gates=["h", "cx"])

        a = expr.Var.new("a", types.Bool())
        good = QuantumCircuit(2, inputs=[a])
        good.store(a, False)
        good.h(0)
        good.cx(0, 1)
        _ = pass_(good)
        self.assertTrue(pass_.property_set["all_gates_in_basis"])

        bad = QuantumCircuit(2, inputs=[a])
        bad.store(a, False)
        bad.x(0)
        bad.cz(0, 1)
        _ = pass_(bad)
        self.assertFalse(pass_.property_set["all_gates_in_basis"])

    def test_store_is_treated_as_builtin_target(self):
        """Test that `Store` is treated as an automatic built-in when given a target."""
        target = Target()
        target.add_instruction(HGate(), {(0,): None, (1,): None})
        target.add_instruction(CXGate(), {(0, 1): None, (1, 0): None})
        pass_ = GatesInBasis(target=target)

        a = expr.Var.new("a", types.Bool())
        good = QuantumCircuit(2, inputs=[a])
        good.store(a, False)
        good.h(0)
        good.cx(0, 1)
        _ = pass_(good)
        self.assertTrue(pass_.property_set["all_gates_in_basis"])

        bad = QuantumCircuit(2, inputs=[a])
        bad.store(a, False)
        bad.x(0)
        bad.cz(0, 1)
        _ = pass_(bad)
        self.assertFalse(pass_.property_set["all_gates_in_basis"])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Backend Properties classes."""

import copy
import datetime
from typing import Any, Iterable, Tuple, Union, Dict
import dateutil.parser

from qiskit.providers.exceptions import BackendPropertyError
from qiskit.utils import deprecate_func
from qiskit.utils.units import apply_prefix

PropertyT = Tuple[Any, datetime.datetime]


class Nduv:
    """Class representing name-date-unit-value

    Attributes:
        date: date.
        name: name.
        unit: unit.
        value: value.
    """

    @deprecate_func(
        since="1.4",
        removal_timeline="in the 2.0 release",
        additional_msg="The models in ``qiskit.providers.models`` and related objects are part "
        "of the deprecated `BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires these representations it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
        stacklevel=2,
    )
    def __init__(self, date, name, unit, value):
        """Initialize a new name-date-unit-value object

        Args:
            date (datetime.datetime): Date field
            name (str): Name field
            unit (str): Nduv unit
            value (float): The value of the Nduv
        """
        self.date = date
        self.name = name
        self.unit = unit
        self.value = value

    @classmethod
    def from_dict(cls, data):
        """Create a new Nduv object from a dictionary.

        Args:
            data (dict): A dictionary representing the Nduv to create.
                         It will be in the same format as output by
                         :func:`to_dict`.

        Returns:
            Nduv: The Nduv from the input dictionary.
        """
        return cls(**data)

    def to_dict(self):
        """Return a dictionary format representation of the object.

        Returns:
            dict: The dictionary form of the Nduv.
        """
        out_dict = {
            "date": self.date,
            "name": self.name,
            "unit": self.unit,
            "value": self.value,
        }
        return out_dict

    def __eq__(self, other):
        if isinstance(other, Nduv):
            if self.to_dict() == other.to_dict():
                return True
        return False

    def __repr__(self):
        return f"Nduv({repr(self.date)}, {self.name}, {self.unit}, {self.value})"


class GateProperties:
    """Class representing a gate's properties

    Attributes:
        qubits: qubits.
        gate: gate.
        parameters: parameters.
    """

    _data = {}

    @deprecate_func(
        since="1.4",
        removal_timeline="in the 2.0 release",
        additional_msg="The models in ``qiskit.providers.models`` and related objects are part "
        "of the deprecated `BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires these representations it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
        stacklevel=2,
    )
    def __init__(self, qubits, gate, parameters, **kwargs):
        """Initialize a new :class:`GateProperties` object

        Args:
            qubits (list): A list of integers representing qubits
            gate (str): The gates name
            parameters (list): List of :class:`Nduv` objects for the
                name-date-unit-value for the gate
            kwargs: Optional additional fields
        """
        self._data = {}
        self.qubits = qubits
        self.gate = gate
        self.parameters = parameters
        self._data.update(kwargs)

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError as ex:
            raise AttributeError(f"Attribute {name} is not defined") from ex

    @classmethod
    def from_dict(cls, data):
        """Create a new Gate object from a dictionary.

        Args:
            data (dict): A dictionary representing the Gate to create.
                         It will be in the same format as output by
                         :func:`to_dict`.

        Returns:
            GateProperties: The Nduv from the input dictionary.
        """
        in_data = {}
        for key, value in data.items():
            if key == "parameters":
                in_data[key] = list(map(Nduv.from_dict, value))
            else:
                in_data[key] = value
        return cls(**in_data)

    def to_dict(self):
        """Return a dictionary format representation of the BackendStatus.

        Returns:
            dict: The dictionary form of the Gate.
        """
        out_dict = {}
        out_dict["qubits"] = self.qubits
        out_dict["gate"] = self.gate
        out_dict["parameters"] = [x.to_dict() for x in self.parameters]
        out_dict.update(self._data)
        return out_dict

    def __eq__(self, other):
        if isinstance(other, GateProperties):
            if self.to_dict() == other.to_dict():
                return True
        return False


# Backwards compatibility.
Gate = GateProperties


class BackendProperties:
    """Class representing backend properties

    This holds backend properties measured by the provider. All properties
    which are provided optionally. These properties may describe qubits, gates,
    or other general properties of the backend.
    """

    _data = {}

    @deprecate_func(
        since="1.2",
        removal_timeline="in the 2.0 release",
        additional_msg="The models in ``qiskit.providers.models`` and related objects are part "
        "of the deprecated `BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
        "workflow requires these representations it likely relies on deprecated functionality and "
        "should be updated to use `BackendV2`.",
        stacklevel=2,
    )
    def __init__(
        self, backend_name, backend_version, last_update_date, qubits, gates, general, **kwargs
    ):
        """Initialize a BackendProperties instance.

        Args:
            backend_name (str): Backend name.
            backend_version (str): Backend version in the form X.Y.Z.
            last_update_date (datetime.datetime or str): Last date/time that a property was
                updated. If specified as a ``str``, it must be in ISO format.
            qubits (list): System qubit parameters as a list of lists of
                           :class:`Nduv` objects
            gates (list): System gate parameters as a list of :class:`GateProperties`
                          objects
            general (list): General parameters as a list of :class:`Nduv`
                            objects
            kwargs: optional additional fields
        """
        self._data = {}
        self.backend_name = backend_name
        self.backend_version = backend_version
        if isinstance(last_update_date, str):
            last_update_date = dateutil.parser.isoparse(last_update_date)
        self.last_update_date = last_update_date
        self.general = general
        self.qubits = qubits
        self.gates = gates

        self._qubits = {}
        for qubit, props in enumerate(qubits):
            formatted_props = {}
            for prop in props:
                value = self._apply_prefix(prop.value, prop.unit)
                formatted_props[prop.name] = (value, prop.date)
                self._qubits[qubit] = formatted_props

        self._gates = {}
        for gate in gates:
            if gate.gate not in self._gates:
                self._gates[gate.gate] = {}
            formatted_props = {}
            for param in gate.parameters:
                value = self._apply_prefix(param.value, param.unit)
                formatted_props[param.name] = (value, param.date)
            self._gates[gate.gate][tuple(gate.qubits)] = formatted_props
        self._data.update(kwargs)

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError as ex:
            raise AttributeError(f"Attribute {name} is not defined") from ex

    @classmethod
    def from_dict(cls, data):
        """Create a new BackendProperties object from a dictionary.

        Args:
            data (dict): A dictionary representing the BackendProperties to create.  It will be in
                the same format as output by :meth:`to_dict`.

        Returns:
            BackendProperties: The BackendProperties from the input dictionary.
        """
        in_data = copy.copy(data)
        backend_name = in_data.pop("backend_name")
        backend_version = in_data.pop("backend_version")
        last_update_date = in_data.pop("last_update_date")
        qubits = []
        for qubit in in_data.pop("qubits"):
            nduvs = []
            for nduv in qubit:
                nduvs.append(Nduv.from_dict(nduv))
            qubits.append(nduvs)
        gates = [GateProperties.from_dict(x) for x in in_data.pop("gates")]
        general = [Nduv.from_dict(x) for x in in_data.pop("general")]

        return cls(
            backend_name, backend_version, last_update_date, qubits, gates, general, **in_data
        )

    def to_dict(self):
        """Return a dictionary format representation of the BackendProperties.

        Returns:
            dict: The dictionary form of the BackendProperties.
        """
        out_dict = {
            "backend_name": self.backend_name,
            "backend_version": self.backend_version,
            "last_update_date": self.last_update_date,
        }
        out_dict["qubits"] = []
        for qubit in self.qubits:
            qubit_props = []
            for item in qubit:
                qubit_props.append(item.to_dict())
            out_dict["qubits"].append(qubit_props)
        out_dict["gates"] = [x.to_dict() for x in self.gates]
        out_dict["general"] = [x.to_dict() for x in self.general]
        out_dict.update(self._data)
        return out_dict

    def __eq__(self, other):
        if isinstance(other, BackendProperties):
            if self.to_dict() == other.to_dict():
                return True
        return False

    def gate_property(
        self,
        gate: str,
        qubits: Union[int, Iterable[int]] = None,
        name: str = None,
    ) -> Union[
        Dict[Tuple[int, ...], Dict[str, PropertyT]],
        Dict[str, PropertyT],
        PropertyT,
    ]:
        """
        Return the property of the given gate.

        Args:
            gate: Name of the gate.
            qubits: The qubit to find the property for.
            name: Optionally used to specify which gate property to return.

        Returns:
            Gate property as a tuple of the value and the time it was measured.

        Raises:
            BackendPropertyError: If the property is not found or name is
                                  specified but qubit is not.
        """
        try:
            result = self._gates[gate]
            if qubits is not None:
                if isinstance(qubits, int):
                    qubits = (qubits,)
                result = result[tuple(qubits)]
                if name:
                    result = result[name]
            elif name:
                raise BackendPropertyError(f"Provide qubits to get {name} of {gate}")
        except KeyError as ex:
            raise BackendPropertyError(f"Could not find the desired property for {gate}") from ex
        return result

    def faulty_qubits(self):
        """Return a list of faulty qubits."""
        faulty = []
        for qubit in self._qubits:
            if not self.is_qubit_operational(qubit):
                faulty.append(qubit)
        return faulty

    def faulty_gates(self):
        """Return a list of faulty gates."""
        faulty = []
        for gate in self.gates:
            if not self.is_gate_operational(gate.gate, gate.qubits):
                faulty.append(gate)
        return faulty

    def is_gate_operational(self, gate: str, qubits: Union[int, Iterable[int]] = None) -> bool:
        """
        Return the operational status of the given gate.

        Args:
            gate: Name of the gate.
            qubits: The qubit to find the operational status for.

        Returns:
            bool: Operational status of the given gate. True if the gate is operational,
            False otherwise.
        """
        properties = self.gate_property(gate, qubits)
        if "operational" in properties:
            return bool(properties["operational"][0])
        return True  # if property operational not existent, then True.

    def gate_error(self, gate: str, qubits: Union[int, Iterable[int]]) -> float:
        """
        Return gate error estimates from backend properties.

        Args:
            gate: The gate for which to get the error.
            qubits: The specific qubits for the gate.

        Returns:
            Gate error of the given gate and qubit(s).
        """
        return self.gate_property(gate, qubits, "gate_error")[0]  # Throw away datetime at index 1

    def gate_length(self, gate: str, qubits: Union[int, Iterable[int]]) -> float:
        """
        Return the duration of the gate in units of seconds.

        Args:
            gate: The gate for which to get the duration.
            qubits: The specific qubits for the gate.

        Returns:
            Gate length of the given gate and qubit(s).
        """
        return self.gate_property(gate, qubits, "gate_length")[0]  # Throw away datetime at index 1

    def qubit_property(
        self,
        qubit: int,
        name: str = None,
    ) -> Union[
        Dict[str, PropertyT],
        PropertyT,
    ]:
        """
        Return the property of the given qubit.

        Args:
            qubit: The property to look for.
            name: Optionally used to specify within the hierarchy which property to return.

        Returns:
            Qubit property as a tuple of the value and the time it was measured.

        Raises:
            BackendPropertyError: If the property is not found.
        """
        try:
            result = self._qubits[qubit]
            if name is not None:
                result = result[name]
        except KeyError as ex:
            formatted_name = "y '" + name + "'" if name else "ies"
            raise BackendPropertyError(
                f"Couldn't find the propert{formatted_name} for qubit {qubit}."
            ) from ex
        return result

    def t1(self, qubit: int) -> float:  # pylint: disable=invalid-name
        """
        Return the T1 time of the given qubit.

        Args:
            qubit: Qubit for which to return the T1 time of.

        Returns:
            T1 time of the given qubit.
        """
        return self.qubit_property(qubit, "T1")[0]  # Throw away datetime at index 1

    def t2(self, qubit: int) -> float:  # pylint: disable=invalid-name
        """
        Return the T2 time of the given qubit.

        Args:
            qubit: Qubit for which to return the T2 time of.

        Returns:
            T2 time of the given qubit.
        """
        return self.qubit_property(qubit, "T2")[0]  # Throw away datetime at index 1

    def frequency(self, qubit: int) -> float:
        """
        Return the frequency of the given qubit.

        Args:
            qubit: Qubit for which to return frequency of.

        Returns:
            Frequency of the given qubit.
        """
        return self.qubit_property(qubit, "frequency")[0]  # Throw away datetime at index 1

    def readout_error(self, qubit: int) -> float:
        """
        Return the readout error of the given qubit.

        Args:
            qubit: Qubit for which to return the readout error of.

        Return:
            Readout error of the given qubit.
        """
        return self.qubit_property(qubit, "readout_error")[0]  # Throw away datetime at index 1

    def readout_length(self, qubit: int) -> float:
        """
        Return the readout length [sec] of the given qubit.

        Args:
            qubit: Qubit for which to return the readout length of.

        Return:
            Readout length of the given qubit.
        """
        return self.qubit_property(qubit, "readout_length")[0]  # Throw away datetime at index 1

    def is_qubit_operational(self, qubit: int) -> bool:
        """
        Return the operational status of the given qubit.

        Args:
            qubit: Qubit for which to return operational status of.

        Returns:
            Operational status of the given qubit.
        """
        properties = self.qubit_property(qubit)
        if "operational" in properties:
            return bool(properties["operational"][0])
        return True  # if property operational not existent, then True.

    def _apply_prefix(self, value: float, unit: str) -> float:
        """
        Given a SI unit prefix and value, apply the prefix to convert to
        standard SI unit.

        Args:
            value: The number to apply prefix to.
            unit: String prefix.

        Returns:
            Converted value.

        Raises:
            BackendPropertyError: If the units aren't recognized.
        """
        try:
            return apply_prefix(value, unit)
        except Exception as ex:
            raise BackendPropertyError(f"Could not understand units: {unit}") from ex



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Tests for Layer2Q implementation.
"""

import unittest
from random import randint
import numpy as np

import qiskit.synthesis.unitary.aqc.fast_gradient.layer as lr
from qiskit.synthesis.unitary.aqc.fast_gradient.pmatrix import PMatrix
from test import QiskitTestCase  # pylint: disable=wrong-import-order
import test.python.synthesis.aqc.fast_gradient.utils_for_testing as tut  # pylint: disable=wrong-import-order


class TestLayer2q(QiskitTestCase):
    """
    Tests for Layer2Q class.
    """

    max_num_qubits = 5  # maximum number of qubits in tests
    num_repeats = 50  # number of repetitions in tests

    def setUp(self):
        super().setUp()
        np.random.seed(0x0696969)

    def test_layer2q_matrix(self):
        """
        Tests: (1) the correctness of Layer2Q matrix construction;
        (2) matrix multiplication interleaved with permutations.
        """

        mat_kind = "complex"
        _eps = 100.0 * np.finfo(float).eps
        max_rel_err = 0.0
        for n in range(2, self.max_num_qubits + 1):

            dim = 2**n
            iden = tut.eye_int(n)
            for j in range(n):
                for k in range(n):
                    if j == k:
                        continue
                    m_mat = tut.rand_matrix(dim=dim, kind=mat_kind)
                    t_mat, g_mat = tut.make_test_matrices4x4(n=n, j=j, k=k, kind=mat_kind)
                    lmat = lr.Layer2Q(num_qubits=n, j=j, k=k, g4x4=g_mat)
                    g2, perm, inv_perm = lmat.get_attr()
                    self.assertTrue(m_mat.dtype == t_mat.dtype == g_mat.dtype == g2.dtype)
                    self.assertTrue(np.all(g_mat == g2))
                    self.assertTrue(np.all(iden[perm].T == iden[inv_perm]))

                    g_mat = np.kron(tut.eye_int(n - 2), g_mat)

                    # T == P^t @ G @ P.
                    err = tut.relative_error(t_mat, iden[perm].T @ g_mat @ iden[perm])
                    self.assertLess(err, _eps, f"err = {err:0.16f}")
                    max_rel_err = max(max_rel_err, err)

                    # Multiplication by permutation matrix of the left can be
                    # replaced by row permutations.
                    tm = t_mat @ m_mat
                    err1 = tut.relative_error(iden[perm].T @ g_mat @ m_mat[perm], tm)
                    err2 = tut.relative_error((g_mat @ m_mat[perm])[inv_perm], tm)

                    # Multiplication by permutation matrix of the right can be
                    # replaced by column permutations.
                    mt = m_mat @ t_mat
                    err3 = tut.relative_error(m_mat @ iden[perm].T @ g_mat @ iden[perm], mt)
                    err4 = tut.relative_error((m_mat[:, perm] @ g_mat)[:, inv_perm], mt)

                    self.assertTrue(
                        err1 < _eps and err2 < _eps and err3 < _eps and err4 < _eps,
                        f"err1 = {err1:f},  err2 = {err2:f},  "
                        f"err3 = {err3:f},  err4 = {err4:f}",
                    )
                    max_rel_err = max(max_rel_err, err1, err2, err3, err4)

    def test_pmatrix_class(self):
        """
        Test the class PMatrix.
        """

        _eps = 100.0 * np.finfo(float).eps
        mat_kind = "complex"
        max_rel_err = 0.0
        for n in range(2, self.max_num_qubits + 1):

            dim = 2**n
            tmp1 = np.ndarray((dim, dim), dtype=np.complex128)
            tmp2 = tmp1.copy()
            for _ in range(self.num_repeats):
                j0 = randint(0, n - 1)
                k0 = (j0 + randint(1, n - 1)) % n
                j1 = randint(0, n - 1)
                k1 = (j1 + randint(1, n - 1)) % n
                j2 = randint(0, n - 1)
                k2 = (j2 + randint(1, n - 1)) % n
                j3 = randint(0, n - 1)
                k3 = (j3 + randint(1, n - 1)) % n
                j4 = randint(0, n - 1)
                k4 = (j4 + randint(1, n - 1)) % n

                t0, g0 = tut.make_test_matrices4x4(n=n, j=j0, k=k0, kind=mat_kind)
                t1, g1 = tut.make_test_matrices4x4(n=n, j=j1, k=k1, kind=mat_kind)
                t2, g2 = tut.make_test_matrices4x4(n=n, j=j2, k=k2, kind=mat_kind)
                t3, g3 = tut.make_test_matrices4x4(n=n, j=j3, k=k3, kind=mat_kind)
                t4, g4 = tut.make_test_matrices4x4(n=n, j=j4, k=k4, kind=mat_kind)

                c0 = lr.Layer2Q(num_qubits=n, j=j0, k=k0, g4x4=g0)
                c1 = lr.Layer2Q(num_qubits=n, j=j1, k=k1, g4x4=g1)
                c2 = lr.Layer2Q(num_qubits=n, j=j2, k=k2, g4x4=g2)
                c3 = lr.Layer2Q(num_qubits=n, j=j3, k=k3, g4x4=g3)
                c4 = lr.Layer2Q(num_qubits=n, j=j4, k=k4, g4x4=g4)

                m_mat = tut.rand_matrix(dim=dim, kind=mat_kind)
                ttmtt = t0 @ t1 @ m_mat @ np.conj(t2).T @ np.conj(t3).T

                pmat = PMatrix(n)
                pmat.set_matrix(m_mat)
                pmat.mul_left_q2(layer=c1, temp_mat=tmp1)
                pmat.mul_left_q2(layer=c0, temp_mat=tmp1)
                pmat.mul_right_q2(layer=c2, temp_mat=tmp1)
                pmat.mul_right_q2(layer=c3, temp_mat=tmp1)
                alt_ttmtt = pmat.finalize(temp_mat=tmp1)

                err1 = tut.relative_error(alt_ttmtt, ttmtt)
                self.assertLess(err1, _eps, f"relative error: {err1:f}")

                prod = np.complex128(np.trace(ttmtt @ t4))
                alt_prod = pmat.product_q2(layer=c4, tmp1=tmp1, tmp2=tmp2)
                err2 = abs(alt_prod - prod) / abs(prod)
                self.assertLess(err2, _eps, f"relative error: {err2:f}")

                max_rel_err = max(max_rel_err, err1, err2)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Object to represent a quantum circuit as a directed acyclic graph (DAG).

The nodes in the graph are either input/output nodes or operation nodes.
The edges correspond to qubits or bits in the circuit. A directed edge
from node A to node B means that the (qu)bit passes from the output of A
to the input of B. The object's methods allow circuits to be constructed,
composed, and modified. Some natural properties like depth can be computed
directly from the graph.
"""

from qiskit._accelerate.circuit import DAGCircuit  # pylint: disable=unused-import



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Remove any swap gates in the circuit by pushing it through into a qubit permutation."""

import logging

from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.layout import Layout
from qiskit._accelerate import elide_permutations as elide_permutations_rs

logger = logging.getLogger(__name__)


class ElidePermutations(TransformationPass):
    r"""Remove permutation operations from a pre-layout circuit

    This pass is intended to be run before a layout (mapping virtual qubits
    to physical qubits) is set during the transpilation pipeline. This
    pass iterates over the :class:`~.DAGCircuit` and when a :class:`~.SwapGate`
    or :class:`~.PermutationGate` are encountered it permutes the virtual qubits in
    the circuit and removes the swap gate. This will effectively remove any
    :class:`~SwapGate`\s or :class:`~PermutationGate` in the circuit prior to running
    layout. If this pass is run after a layout has been set it will become a no-op
    (and log a warning) as this optimization is not sound after physical qubits are
    selected and there are connectivity constraints to adhere to.

    For tracking purposes this pass sets 3 values in the property set if there
    are any :class:`~.SwapGate` or :class:`~.PermutationGate` objects in the circuit
    and the pass isn't a no-op.

    * ``original_layout``: The trivial :class:`~.Layout` for the input to this pass being run
    * ``original_qubit_indices``: The mapping of qubit objects to positional indices for the state
        of the circuit as input to this pass.
    * ``virtual_permutation_layout``: A :class:`~.Layout` object mapping input qubits to the output
        state after eliding permutations.

    These three properties are needed for the transpiler to track the permutations in the out
    :attr:`.QuantumCircuit.layout` attribute. The elision of permutations is equivalent to a
    ``final_layout`` set by routing and all three of these attributes are needed in the case
    """

    def run(self, dag):
        """Run the ElidePermutations pass on ``dag``.

        Args:
            dag (DAGCircuit): the DAG to be optimized.

        Returns:
            DAGCircuit: the optimized DAG.
        """
        if self.property_set["layout"] is not None:
            logger.warning(
                "ElidePermutations is not valid after a layout has been set. This indicates "
                "an invalid pass manager construction."
            )
            return dag

        result = elide_permutations_rs.run(dag)

        # If the pass did not do anything, the result is None
        if result is None:
            return dag

        # Otherwise, the result is a pair consisting of the rewritten DAGCircuit and the
        # qubit mapping.
        (new_dag, qubit_mapping) = result

        input_qubit_mapping = {qubit: index for index, qubit in enumerate(dag.qubits)}
        self.property_set["original_layout"] = Layout(input_qubit_mapping)
        if self.property_set["original_qubit_indices"] is None:
            self.property_set["original_qubit_indices"] = input_qubit_mapping

        new_layout = Layout({dag.qubits[out]: idx for idx, out in enumerate(qubit_mapping)})
        if current_layout := self.property_set["virtual_permutation_layout"]:
            self.property_set["virtual_permutation_layout"] = new_layout.compose(
                current_layout.inverse(dag.qubits, dag.qubits), dag.qubits
            )
        else:
            self.property_set["virtual_permutation_layout"] = new_layout
        return new_dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""Randomized tests of transpiler circuit equivalence.

This test can be optionally configured (e.g. by CI) via the
following env vars:

QISKIT_RANDOMIZED_TEST_LAYOUT_METHODS

    A space-delimited list of layout method names from which the
    randomizer should pick the layout method. Defaults to all
    available built-in methods if unspecified.

QISKIT_RANDOMIZED_TEST_ROUTING_METHODS

    A space-delimited list of routing method names from which the
    randomizer should pick the routing method. Defaults to all
    available built-in methods if unspecified.

QISKIT_RANDOMIZED_TEST_SCHEDULING_METHODS

    A space-delimited list of scheduling method names from which the
    randomizer should pick the scheduling method. Defaults to all
    available built-in methods if unspecified.

QISKIT_RANDOMIZED_TEST_BACKEND_NEEDS_DURATIONS

    A boolean value (e.g. "true", "Y", etc.) which, when true, forces
    the randomizer to pick a backend which fully supports scheduling
    (i.e. has fully specified duration info). Defaults to False.

QISKIT_RANDOMIZED_TEST_ALLOW_BARRIERS

    A boolean value (e.g. "true", "Y", etc.) which, when false,
    prevents the randomizer from emitting barrier instructions.
    Defaults to True.
"""

import os
import warnings

from test.utils.base import dicts_almost_equal

from math import pi

from hypothesis import assume, settings, HealthCheck
from hypothesis.stateful import multiple, rule, precondition, invariant
from hypothesis.stateful import Bundle, RuleBasedStateMachine
import hypothesis.strategies as st

from qiskit import transpile, qasm2
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit.circuit import Measure, Reset, Gate, Barrier
from qiskit.providers.fake_provider import (
    Fake5QV1,
    Fake20QV1,
    Fake7QPulseV1,
    Fake27QPulseV1,
    Fake127QPulseV1,
)

# pylint: disable=wildcard-import,unused-wildcard-import
from qiskit.circuit.library.standard_gates import *

from qiskit_aer import Aer  # pylint: disable=wrong-import-order

default_profile = "transpiler_equivalence"
settings.register_profile(
    default_profile,
    report_multiple_bugs=False,
    max_examples=200,
    deadline=None,
    suppress_health_check=[HealthCheck.filter_too_much],
)

settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", default_profile))

BASE_INSTRUCTIONS = {
    # Key is (n_qubits, n_clbits, n_params).  All gates here should be directly known by Aer so they
    # can be simulated without an initial transpile (whether that's via `execute` or not).
    (1, 0, 0): [HGate, IGate, SGate, SdgGate, TGate, TdgGate, XGate, YGate, ZGate, Reset],
    (2, 0, 0): [CXGate, CYGate, CZGate, SwapGate],
    (3, 0, 0): [CCXGate, CSwapGate],
    (1, 0, 1): [PhaseGate, RXGate, RYGate, RZGate],
    (1, 0, 3): [UGate],
    (2, 0, 1): [RZZGate, CPhaseGate],
    (2, 0, 4): [CUGate],
    (1, 1, 0): [Measure],
}
variadic_gates = [Barrier]


def _strtobool(s):
    return s.lower() in ("y", "yes", "t", "true", "on", "1")


if not _strtobool(os.getenv("QISKIT_RANDOMIZED_TEST_ALLOW_BARRIERS", "True")):
    variadic_gates.remove(Barrier)


def _getenv_list(var_name):
    value = os.getenv(var_name)
    return None if value is None else value.split()


# Note: a value of `None` for any of the following methods means that
# the selected pass manager gets to choose. However, to avoid complexity,
# its not possible to specify `None` when overriding these with environment
# variables. Really, `None` is useful only for testing Terra's pass managers,
# and if you're overriding these, your goal is probably to test a specific
# pass or set of passes instead.
layout_methods = _getenv_list("QISKIT_RANDOMIZED_TEST_LAYOUT_METHODS") or [
    None,
    "trivial",
    "dense",
    "sabre",
]
routing_methods = _getenv_list("QISKIT_RANDOMIZED_TEST_ROUTING_METHODS") or [
    None,
    "basic",
    "stochastic",
    "lookahead",
    "sabre",
]
scheduling_methods = _getenv_list("QISKIT_RANDOMIZED_TEST_SCHEDULING_METHODS") or [
    None,
    "alap",
    "asap",
]

backend_needs_durations = _strtobool(
    os.getenv("QISKIT_RANDOMIZED_TEST_BACKEND_NEEDS_DURATIONS", "False")
)


def _fully_supports_scheduling(backend):
    """Checks if backend is not in the set of backends known not to have specified gate durations."""
    return not isinstance(
        backend,
        (Fake20QV1, Fake5QV1),
    )


mock_backends = [Fake5QV1(), Fake20QV1(), Fake7QPulseV1(), Fake27QPulseV1(), Fake127QPulseV1()]

mock_backends_with_scheduling = [b for b in mock_backends if _fully_supports_scheduling(b)]


@st.composite
def transpiler_conf(draw):
    """Composite search strategy to pick a valid transpiler config."""
    all_backends = st.one_of(st.none(), st.sampled_from(mock_backends))
    scheduling_backends = st.sampled_from(mock_backends_with_scheduling)
    scheduling_method = draw(st.sampled_from(scheduling_methods))
    backend = (
        draw(scheduling_backends)
        if scheduling_method or backend_needs_durations
        else draw(all_backends)
    )
    return {
        "backend": backend,
        "optimization_level": draw(st.integers(min_value=0, max_value=3)),
        "layout_method": draw(st.sampled_from(layout_methods)),
        "routing_method": draw(st.sampled_from(routing_methods)),
        "scheduling_method": scheduling_method,
        "seed_transpiler": draw(st.integers(min_value=0, max_value=1_000_000)),
    }


class QCircuitMachine(RuleBasedStateMachine):
    """Build a Hypothesis rule based state machine for constructing, transpiling
    and simulating a series of random QuantumCircuits.

    Build circuits with up to QISKIT_RANDOM_QUBITS qubits, apply a random
    selection of gates from qiskit.circuit.library with randomly selected
    qargs, cargs, and parameters. At random intervals, transpile the circuit for
    a random backend with a random optimization level and simulate both the
    initial and the transpiled circuits to verify that their counts are the
    same.
    """

    qubits = Bundle("qubits")
    clbits = Bundle("clbits")

    backend = Aer.get_backend("aer_simulator")
    max_qubits = int(backend.configuration().n_qubits / 2)

    # Limit reg generation for more interesting circuits
    max_qregs = 3
    max_cregs = 3

    def __init__(self):
        super().__init__()
        self.qc = QuantumCircuit()
        self.enable_variadic = bool(variadic_gates)

    @precondition(lambda self: len(self.qc.qubits) < self.max_qubits)
    @precondition(lambda self: len(self.qc.qregs) < self.max_qregs)
    @rule(target=qubits, n=st.integers(min_value=1, max_value=max_qubits))
    def add_qreg(self, n):
        """Adds a new variable sized qreg to the circuit, up to max_qubits."""
        n = min(n, self.max_qubits - len(self.qc.qubits))
        qreg = QuantumRegister(n)
        self.qc.add_register(qreg)
        return multiple(*list(qreg))

    @precondition(lambda self: len(self.qc.cregs) < self.max_cregs)
    @rule(target=clbits, n=st.integers(1, 5))
    def add_creg(self, n):
        """Add a new variable sized creg to the circuit."""
        creg = ClassicalRegister(n)
        self.qc.add_register(creg)
        return multiple(*list(creg))

    # Gates of various shapes

    @precondition(lambda self: self.qc.num_qubits > 0 and self.qc.num_clbits > 0)
    @rule(n_arguments=st.sampled_from(sorted(BASE_INSTRUCTIONS.keys())), data=st.data())
    def add_gate(self, n_arguments, data):
        """Append a random fixed gate to the circuit."""
        n_qubits, n_clbits, n_params = n_arguments
        gate_class = data.draw(st.sampled_from(BASE_INSTRUCTIONS[n_qubits, n_clbits, n_params]))
        qubits = data.draw(st.lists(self.qubits, min_size=n_qubits, max_size=n_qubits, unique=True))
        clbits = data.draw(st.lists(self.clbits, min_size=n_clbits, max_size=n_clbits, unique=True))
        params = data.draw(
            st.lists(
                st.floats(
                    allow_nan=False, allow_infinity=False, min_value=-10 * pi, max_value=10 * pi
                ),
                min_size=n_params,
                max_size=n_params,
            )
        )
        self.qc.append(gate_class(*params), qubits, clbits)

    @precondition(lambda self: self.enable_variadic)
    @rule(gate=st.sampled_from(variadic_gates), qargs=st.lists(qubits, min_size=1, unique=True))
    def add_variQ_gate(self, gate, qargs):
        """Append a gate with a variable number of qargs."""
        self.qc.append(gate(len(qargs)), qargs)

    @precondition(lambda self: len(self.qc.data) > 0)
    @rule(carg=clbits, data=st.data())
    def add_c_if_last_gate(self, carg, data):
        """Modify the last gate to be conditional on a classical register."""
        creg = self.qc.find_bit(carg).registers[0][0]
        val = data.draw(st.integers(min_value=0, max_value=2 ** len(creg) - 1))

        last_gate = self.qc.data[-1]

        # Conditional instructions are not supported
        assume(isinstance(last_gate.operation, Gate))

        last_gate.operation.c_if(creg, val)

    # Properties to check

    @invariant()
    def qasm(self):
        """After each circuit operation, it should be possible to build QASM."""
        qasm2.dumps(self.qc)

    @precondition(lambda self: any(isinstance(d.operation, Measure) for d in self.qc.data))
    @rule(kwargs=transpiler_conf())
    def equivalent_transpile(self, kwargs):
        """Simulate, transpile and simulate the present circuit. Verify that the
        counts are not significantly different before and after transpilation.

        """
        assume(
            kwargs["backend"] is None
            or kwargs["backend"].configuration().n_qubits >= len(self.qc.qubits)
        )

        call = (
            "transpile(qc, "
            + ", ".join(f"{key:s}={value!r}" for key, value in kwargs.items() if value is not None)
            + ")"
        )
        print(f"Evaluating {call} for:\n{qasm2.dumps(self.qc)}")

        shots = 4096

        # Note that there's no transpilation here, which is why the gates are limited to only ones
        # that Aer supports natively.
        with warnings.catch_warnings():
            # Safe to remove once https://github.com/Qiskit/qiskit-aer/pull/2179 is in a release version
            # of Aer.
            warnings.filterwarnings(
                "default",
                category=DeprecationWarning,
                module="qiskit_aer",
                message="Treating CircuitInstruction as an iterable",
            )
            aer_counts = self.backend.run(self.qc, shots=shots).result().get_counts()

        try:
            xpiled_qc = transpile(self.qc, **kwargs)
        except Exception as e:
            failed_qasm = (
                f"Exception caught during transpilation of circuit: \n{qasm2.dumps(self.qc)}"
            )
            raise RuntimeError(failed_qasm) from e

        xpiled_aer_counts = self.backend.run(xpiled_qc, shots=shots).result().get_counts()

        count_differences = dicts_almost_equal(aer_counts, xpiled_aer_counts, 0.05 * shots)

        assert count_differences == "", (
            f"Counts not equivalent: {count_differences}\nFailing QASM Input:\n"
            f"{qasm2.dumps(self.qc)}\n\nFailing QASM Output:\n{qasm2.dumps(xpiled_qc)}"
        )


TestQuantumCircuit = QCircuitMachine.TestCase



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-function-docstring, missing-module-docstring

import unittest

from qiskit import QuantumCircuit
from qiskit.providers.basic_provider import BasicSimulator
import qiskit.circuit.library.standard_gates as lib
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestStandardGates(QiskitTestCase):
    """Standard gates support in BasicSimulator, up to 3 qubits"""

    def setUp(self):
        super().setUp()
        self.seed = 43
        self.shots = 1
        self.circuit = QuantumCircuit(4)

    def test_barrier(self):
        self.circuit.barrier(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_barrier_none(self):
        self.circuit.barrier()
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_unitary(self):
        matrix = [[0, 0, 0, 1], [0, 0, 1, 0], [1, 0, 0, 0], [0, 1, 0, 0]]
        self.circuit.unitary(matrix, [0, 1])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_u(self):
        self.circuit.u(0.5, 1.5, 1.5, 0)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_u1(self):
        self.circuit.append(lib.U1Gate(0.5), [1])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_u2(self):
        self.circuit.append(lib.U2Gate(0.5, 0.5), [1])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_u3(self):
        self.circuit.append(lib.U3Gate(0.5, 0.5, 0.5), [1])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_ccx(self):
        self.circuit.ccx(0, 1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_ccz(self):
        self.circuit.ccz(0, 1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_ch(self):
        self.circuit.ch(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cp(self):
        self.circuit.cp(0, 0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_crx(self):
        self.circuit.crx(1, 0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cry(self):
        self.circuit.cry(1, 0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_crz(self):
        self.circuit.crz(1, 0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cswap(self):
        self.circuit.cswap(0, 1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cu1(self):
        self.circuit.append(lib.CU1Gate(1), [1, 2])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cu3(self):
        self.circuit.append(lib.CU3Gate(1, 2, 3), [1, 2])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cx(self):
        self.circuit.cx(1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_ecr(self):
        self.circuit.ecr(1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cy(self):
        self.circuit.cy(1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cz(self):
        self.circuit.cz(1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_h(self):
        self.circuit.h(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_id(self):
        self.circuit.id(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_rx(self):
        self.circuit.rx(1, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_ry(self):
        self.circuit.ry(1, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_rz(self):
        self.circuit.rz(1, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_rxx(self):
        self.circuit.rxx(1, 1, 0)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_rzx(self):
        self.circuit.rzx(1, 1, 0)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_ryy(self):
        self.circuit.ryy(1, 1, 0)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_rzz(self):
        self.circuit.rzz(1, 1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_s(self):
        self.circuit.s(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_sdg(self):
        self.circuit.sdg(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_sx(self):
        self.circuit.sx(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_sxdg(self):
        self.circuit.sxdg(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_swap(self):
        self.circuit.swap(1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_iswap(self):
        self.circuit.iswap(1, 0)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_p(self):
        self.circuit.p(1, 0)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_r(self):
        self.circuit.r(0.5, 0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_t(self):
        self.circuit.t(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_tdg(self):
        self.circuit.tdg(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_x(self):
        self.circuit.x(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_y(self):
        self.circuit.y(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_z(self):
        self.circuit.z(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cs(self):
        self.circuit.cs(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_csdg(self):
        self.circuit.csdg(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_csx(self):
        self.circuit.csx(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_cu(self):
        self.circuit.cu(0.5, 0.5, 0.5, 0.5, 0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_dcx(self):
        self.circuit.dcx(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_delay(self):
        self.circuit.delay(0, 1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_reset(self):
        self.circuit.reset(1)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_rcx(self):
        self.circuit.rccx(0, 1, 2)
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_global_phase(self):
        qc = self.circuit
        qc.append(lib.GlobalPhaseGate(0.1), [])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_xx_minus_yy(self):
        self.circuit.append(lib.XXMinusYYGate(0.1, 0.2), [0, 1])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)

    def test_xx_plus_yy(self):
        self.circuit.append(lib.XXPlusYYGate(0.1, 0.2), [0, 1])
        self.circuit.measure_all()
        result = (
            BasicSimulator().run(self.circuit, shots=self.shots, seed_simulator=self.seed).result()
        )
        self.assertEqual(result.success, True)


class TestStandardGatesTarget(QiskitTestCase):
    """Standard gates, up to 3 qubits, as a target"""

    def test_target(self):
        target = BasicSimulator().target
        expected = {
            "cz",
            "u3",
            "p",
            "cswap",
            "z",
            "cu1",
            "ecr",
            "reset",
            "ch",
            "cy",
            "dcx",
            "crx",
            "sx",
            "unitary",
            "csdg",
            "rzz",
            "measure",
            "swap",
            "csx",
            "y",
            "s",
            "xx_plus_yy",
            "cs",
            "h",
            "t",
            "u",
            "rxx",
            "cu",
            "rzx",
            "ry",
            "rx",
            "cu3",
            "tdg",
            "u2",
            "xx_minus_yy",
            "global_phase",
            "u1",
            "id",
            "cx",
            "cp",
            "rz",
            "sxdg",
            "x",
            "ryy",
            "sdg",
            "ccz",
            "delay",
            "crz",
            "iswap",
            "ccx",
            "cry",
            "rccx",
            "r",
        }
        self.assertEqual(set(target.operation_names), expected)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
A definition of the approximate circuit compilation optimization problem based on CNOT unit
definition.
"""
from __future__ import annotations
import typing
from abc import ABC

import numpy as np
from numpy import linalg as la

from .approximate import ApproximatingObjective
from .elementary_operations import ry_matrix, rz_matrix, place_unitary, place_cnot, rx_matrix


class CNOTUnitObjective(ApproximatingObjective, ABC):
    """
    A base class for a problem definition based on CNOT unit. This class may have different
    subclasses for objective and gradient computations.
    """

    def __init__(self, num_qubits: int, cnots: np.ndarray) -> None:
        """
        Args:
            num_qubits: number of qubits.
            cnots: a CNOT structure to be used in the optimization procedure.
        """
        super().__init__()
        self._num_qubits = num_qubits
        self._cnots = cnots
        self._num_cnots = cnots.shape[1]

    @property
    def num_cnots(self):
        """
        Returns:
            A number of CNOT units to be used by the approximate circuit.
        """
        return self._num_cnots

    @property
    def num_thetas(self):
        """
        Returns:
            Number of parameters (angles) of rotation gates in this circuit.
        """
        return 3 * self._num_qubits + 4 * self._num_cnots


class DefaultCNOTUnitObjective(CNOTUnitObjective):
    """A naive implementation of the objective function based on CNOT units."""

    def __init__(self, num_qubits: int, cnots: np.ndarray) -> None:
        """
        Args:
            num_qubits: number of qubits.
            cnots: a CNOT structure to be used in the optimization procedure.
        """
        super().__init__(num_qubits, cnots)

        # last objective computations to be re-used by gradient
        self._last_thetas: np.ndarray | None = None
        self._cnot_right_collection: np.ndarray | None = None
        self._cnot_left_collection: np.ndarray | None = None
        self._rotation_matrix: int | np.ndarray | None = None
        self._cnot_matrix: np.ndarray | None = None

    def objective(self, param_values: np.ndarray) -> typing.SupportsFloat:
        # rename parameters just to make shorter and make use of our dictionary
        thetas = param_values
        n = self._num_qubits
        d = int(2**n)
        cnots = self._cnots
        num_cnots = self.num_cnots

        # to save intermediate computations we define the following matrices
        # this is the collection of cnot unit matrices ordered from left to
        # right as in the circuit, not matrix product
        cnot_unit_collection = np.zeros((d, d * num_cnots), dtype=complex)
        # this is the collection of matrix products of the cnot units up
        # to the given position from the right of the circuit
        cnot_right_collection = np.zeros((d, d * num_cnots), dtype=complex)
        # this is the collection of matrix products of the cnot units up
        # to the given position from the left of the circuit
        cnot_left_collection = np.zeros((d, d * num_cnots), dtype=complex)
        # first, we construct each cnot unit matrix
        for cnot_index in range(num_cnots):
            theta_index = 4 * cnot_index

            # cnot qubit indices for the cnot unit identified by cnot_index
            q1 = int(cnots[0, cnot_index])
            q2 = int(cnots[1, cnot_index])

            # rotations that are applied on the q1 qubit
            ry1 = ry_matrix(thetas[0 + theta_index])
            rz1 = rz_matrix(thetas[1 + theta_index])

            # rotations that are applied on the q2 qubit
            ry2 = ry_matrix(thetas[2 + theta_index])
            rx2 = rx_matrix(thetas[3 + theta_index])

            # combine the rotations on qubits q1 and q2
            single_q1 = np.dot(rz1, ry1)
            single_q2 = np.dot(rx2, ry2)

            # we place single qubit matrices at the corresponding locations in the (2^n, 2^n) matrix
            full_q1 = place_unitary(single_q1, n, q1)
            full_q2 = place_unitary(single_q2, n, q2)

            # we place a cnot matrix at the qubits q1 and q2 in the full matrix
            cnot_q1q2 = place_cnot(n, q1, q2)

            # compute the cnot unit matrix and store in cnot_unit_collection
            cnot_unit_collection[:, d * cnot_index : d * (cnot_index + 1)] = la.multi_dot(
                [full_q2, full_q1, cnot_q1q2]
            )

        # this is the matrix corresponding to the intermediate matrix products
        # it will end up being the matrix product of all the cnot unit matrices
        # first we multiply from the right-hand side of the circuit
        cnot_matrix = np.eye(d)
        for cnot_index in range(num_cnots - 1, -1, -1):
            cnot_matrix = np.dot(
                cnot_matrix, cnot_unit_collection[:, d * cnot_index : d * (cnot_index + 1)]
            )
            cnot_right_collection[:, d * cnot_index : d * (cnot_index + 1)] = cnot_matrix
        # now we multiply from the left-hand side of the circuit
        cnot_matrix = np.eye(d)
        for cnot_index in range(num_cnots):
            cnot_matrix = np.dot(
                cnot_unit_collection[:, d * cnot_index : d * (cnot_index + 1)], cnot_matrix
            )
            cnot_left_collection[:, d * cnot_index : d * (cnot_index + 1)] = cnot_matrix

        # this is the matrix corresponding to the initial rotations
        # we start with 1 and kronecker product each qubit's rotations
        rotation_matrix: int | np.ndarray = 1
        for q in range(n):
            theta_index = 4 * num_cnots + 3 * q
            rz0 = rz_matrix(thetas[0 + theta_index])
            ry1 = ry_matrix(thetas[1 + theta_index])
            rz2 = rz_matrix(thetas[2 + theta_index])
            rotation_matrix = np.kron(rotation_matrix, la.multi_dot([rz0, ry1, rz2]))

        # the matrix corresponding to the full circuit is the cnot part and
        # rotation part multiplied together
        circuit_matrix = np.dot(cnot_matrix, rotation_matrix)

        # compute error
        error = 0.5 * (la.norm(circuit_matrix - self._target_matrix, "fro") ** 2)

        # cache computations for gradient
        self._last_thetas = thetas
        self._cnot_left_collection = cnot_left_collection
        self._cnot_right_collection = cnot_right_collection
        self._rotation_matrix = rotation_matrix
        self._cnot_matrix = cnot_matrix

        return error

    def gradient(self, param_values: np.ndarray) -> np.ndarray:
        # just to make shorter
        thetas = param_values
        # if given thetas are the same as used at the previous objective computations, then
        # we re-use computations, otherwise we have to re-compute objective
        if not np.all(np.isclose(thetas, self._last_thetas)):
            self.objective(thetas)

        # the partial derivative of the circuit with respect to an angle
        # is the same circuit with the corresponding pauli gate, multiplied
        # by a global phase of -1j / 2, next to the rotation gate (it commutes)
        pauli_x = np.multiply(-1j / 2, np.asarray([[0, 1], [1, 0]]))
        pauli_y = np.multiply(-1j / 2, np.asarray([[0, -1j], [1j, 0]]))
        pauli_z = np.multiply(-1j / 2, np.asarray([[1, 0], [0, -1]]))

        n = self._num_qubits
        d = int(2**n)
        cnots = self._cnots
        num_cnots = self.num_cnots

        # the partial derivative of the cost function is -Re<V',U>
        # where V' is the partial derivative of the circuit
        # first we compute the partial derivatives in the cnot part
        der = np.zeros(4 * num_cnots + 3 * n)
        for cnot_index in range(num_cnots):
            theta_index = 4 * cnot_index

            # cnot qubit indices for the cnot unit identified by cnot_index
            q1 = int(cnots[0, cnot_index])
            q2 = int(cnots[1, cnot_index])

            # rotations that are applied on the q1 qubit
            ry1 = ry_matrix(thetas[0 + theta_index])
            rz1 = rz_matrix(thetas[1 + theta_index])

            # rotations that are applied on the q2 qubit
            ry2 = ry_matrix(thetas[2 + theta_index])
            rx2 = rx_matrix(thetas[3 + theta_index])

            # combine the rotations on qubits q1 and q2
            # note we have to insert an extra pauli gate to take the derivative
            # of the appropriate rotation gate
            for i in range(4):
                if i == 0:
                    single_q1 = la.multi_dot([rz1, pauli_y, ry1])
                    single_q2 = np.dot(rx2, ry2)
                elif i == 1:
                    single_q1 = la.multi_dot([pauli_z, rz1, ry1])
                    single_q2 = np.dot(rx2, ry2)
                elif i == 2:
                    single_q1 = np.dot(rz1, ry1)
                    single_q2 = la.multi_dot([rx2, pauli_y, ry2])
                else:
                    single_q1 = np.dot(rz1, ry1)
                    single_q2 = la.multi_dot([pauli_x, rx2, ry2])

                # we place single qubit matrices at the corresponding locations in
                # the (2^n, 2^n) matrix
                full_q1 = place_unitary(single_q1, n, q1)
                full_q2 = place_unitary(single_q2, n, q2)

                # we place a cnot matrix at the qubits q1 and q2 in the full matrix
                cnot_q1q2 = place_cnot(n, q1, q2)

                # partial derivative of that particular cnot unit, size of (2^n, 2^n)
                der_cnot_unit = la.multi_dot([full_q2, full_q1, cnot_q1q2])
                # der_cnot_unit is multiplied by the matrix product of cnot units to the left
                # of it (if there are any) and to the right of it (if there are any)
                if cnot_index == 0:
                    der_cnot_matrix = np.dot(
                        self._cnot_right_collection[:, d : 2 * d],
                        der_cnot_unit,
                    )
                elif num_cnots - 1 == cnot_index:
                    der_cnot_matrix = np.dot(
                        der_cnot_unit,
                        self._cnot_left_collection[:, d * (num_cnots - 2) : d * (num_cnots - 1)],
                    )
                else:
                    der_cnot_matrix = la.multi_dot(
                        [
                            self._cnot_right_collection[
                                :, d * (cnot_index + 1) : d * (cnot_index + 2)
                            ],
                            der_cnot_unit,
                            self._cnot_left_collection[:, d * (cnot_index - 1) : d * cnot_index],
                        ]
                    )

                # the matrix corresponding to the full circuit partial derivative
                # is the partial derivative of the cnot part multiplied by the usual
                # rotation part
                der_circuit_matrix = np.dot(der_cnot_matrix, self._rotation_matrix)
                # we compute the partial derivative of the cost function
                der[i + theta_index] = -np.real(
                    np.trace(np.dot(der_circuit_matrix.conj().T, self._target_matrix))
                )

        # now we compute the partial derivatives in the rotation part
        # we start with 1 and kronecker product each qubit's rotations
        for i in range(3 * n):
            der_rotation_matrix: int | np.ndarray = 1
            for q in range(n):
                theta_index = 4 * num_cnots + 3 * q
                rz0 = rz_matrix(thetas[0 + theta_index])
                ry1 = ry_matrix(thetas[1 + theta_index])
                rz2 = rz_matrix(thetas[2 + theta_index])
                # for the appropriate rotation gate that we are taking
                # the partial derivative of, we have to insert the
                # corresponding pauli matrix
                if i - 3 * q == 0:
                    rz0 = np.dot(pauli_z, rz0)
                elif i - 3 * q == 1:
                    ry1 = np.dot(pauli_y, ry1)
                elif i - 3 * q == 2:
                    rz2 = np.dot(pauli_z, rz2)
                der_rotation_matrix = np.kron(der_rotation_matrix, la.multi_dot([rz0, ry1, rz2]))

            # the matrix corresponding to the full circuit partial derivative
            # is the usual cnot part multiplied by the partial derivative of
            # the rotation part
            der_circuit_matrix = np.dot(self._cnot_matrix, der_rotation_matrix)
            # we compute the partial derivative of the cost function
            der[4 * num_cnots + i] = -np.real(
                np.trace(np.dot(der_circuit_matrix.conj().T, self._target_matrix))
            )

        return der



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""
Chi-matrix representation of a Quantum Channel.
"""

from __future__ import annotations
import copy as _copy
import math
from typing import TYPE_CHECKING

import numpy as np

from qiskit import _numpy_compat
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.instruction import Instruction
from qiskit.exceptions import QiskitError
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel
from qiskit.quantum_info.operators.channel.choi import Choi
from qiskit.quantum_info.operators.channel.superop import SuperOp
from qiskit.quantum_info.operators.channel.transformations import _to_chi
from qiskit.quantum_info.operators.mixins import generate_apidocs
from qiskit.quantum_info.operators.base_operator import BaseOperator

if TYPE_CHECKING:
    from qiskit import circuit


class Chi(QuantumChannel):
    r"""Pauli basis Chi-matrix representation of a quantum channel.

    The Chi-matrix representation of an :math:`n`-qubit quantum channel
    :math:`\mathcal{E}` is a matrix :math:`\chi` such that the evolution of a
    :class:`~qiskit.quantum_info.DensityMatrix` :math:`\rho` is given by

    .. math::

        \mathcal{E}(ρ) = \frac{1}{2^n} \sum_{i, j} \chi_{i,j} P_i ρ P_j

    where :math:`[P_0, P_1, ..., P_{4^{n}-1}]` is the :math:`n`-qubit Pauli basis in
    lexicographic order. It is related to the :class:`Choi` representation by a change
    of basis of the Choi-matrix into the Pauli basis. The :math:`\frac{1}{2^n}`
    in the definition above is a normalization factor that arises from scaling the
    Pauli basis to make it orthonormal.

    See reference [1] for further details.

    References:
        1. C.J. Wood, J.D. Biamonte, D.G. Cory, *Tensor networks and graphical calculus
           for open quantum systems*, Quant. Inf. Comp. 15, 0579-0811 (2015).
           `arXiv:1111.6950 [quant-ph] <https://arxiv.org/abs/1111.6950>`_
    """

    def __init__(
        self,
        data: QuantumCircuit | circuit.instruction.Instruction | BaseOperator | np.ndarray,
        input_dims: int | tuple | None = None,
        output_dims: int | tuple | None = None,
    ):
        """Initialize a quantum channel Chi-matrix operator.

        Args:
            data: data to initialize superoperator.
            input_dims: the input subsystem dimensions.
            output_dims: the output subsystem dimensions.

        Raises:
            QiskitError: if input data is not an N-qubit channel or
                         cannot be initialized as a Chi-matrix.

        Additional Information:
            If the input or output dimensions are None, they will be
            automatically determined from the input data. The Chi matrix
            representation is only valid for N-qubit channels.
        """
        # If the input is a raw list or matrix we assume that it is
        # already a Chi matrix.
        if isinstance(data, (list, np.ndarray)):
            # Initialize from raw numpy or list matrix.
            chi_mat = np.asarray(data, dtype=complex)
            # Determine input and output dimensions
            dim_l, dim_r = chi_mat.shape
            if dim_l != dim_r:
                raise QiskitError("Invalid Chi-matrix input.")
            if input_dims:
                input_dim = np.prod(input_dims)
            if output_dims:
                output_dim = np.prod(input_dims)
            if output_dims is None and input_dims is None:
                output_dim = int(math.sqrt(dim_l))
                input_dim = dim_l // output_dim
            elif input_dims is None:
                input_dim = dim_l // output_dim
            elif output_dims is None:
                output_dim = dim_l // input_dim
            # Check dimensions
            if input_dim * output_dim != dim_l:
                raise QiskitError("Invalid shape for Chi-matrix input.")
        else:
            # Otherwise we initialize by conversion from another Qiskit
            # object into the QuantumChannel.
            if isinstance(data, (QuantumCircuit, Instruction)):
                # If the input is a Terra QuantumCircuit or Instruction we
                # convert it to a SuperOp
                data = SuperOp._init_instruction(data)
            else:
                # We use the QuantumChannel init transform to initialize
                # other objects into a QuantumChannel or Operator object.
                data = self._init_transformer(data)
            input_dim, output_dim = data.dim
            # Now that the input is an operator we convert it to a Chi object
            rep = getattr(data, "_channel_rep", "Operator")
            chi_mat = _to_chi(rep, data._data, input_dim, output_dim)
            if input_dims is None:
                input_dims = data.input_dims()
            if output_dims is None:
                output_dims = data.output_dims()
        # Check input is N-qubit channel
        num_qubits = int(math.log2(input_dim))
        if 2**num_qubits != input_dim or input_dim != output_dim:
            raise QiskitError("Input is not an n-qubit Chi matrix.")
        super().__init__(chi_mat, num_qubits=num_qubits)

    def __array__(self, dtype=None, copy=_numpy_compat.COPY_ONLY_IF_NEEDED):
        dtype = self.data.dtype
        return np.array(self.data, dtype=dtype, copy=copy)

    @property
    def _bipartite_shape(self):
        """Return the shape for bipartite matrix"""
        return (self._input_dim, self._output_dim, self._input_dim, self._output_dim)

    def _evolve(self, state, qargs=None):
        return SuperOp(self)._evolve(state, qargs)

    # ---------------------------------------------------------------------
    # BaseOperator methods
    # ---------------------------------------------------------------------

    def conjugate(self):
        # Since conjugation is basis dependent we transform
        # to the Choi representation to compute the
        # conjugate channel
        return Chi(Choi(self).conjugate())

    def transpose(self):
        return Chi(Choi(self).transpose())

    def adjoint(self):
        return Chi(Choi(self).adjoint())

    def compose(self, other: Chi, qargs: list | None = None, front: bool = False) -> Chi:
        if qargs is None:
            qargs = getattr(other, "qargs", None)
        if qargs is not None:
            return Chi(SuperOp(self).compose(other, qargs=qargs, front=front))
        # If no qargs we compose via Choi representation to avoid an additional
        # representation conversion to SuperOp and then convert back to Chi
        return Chi(Choi(self).compose(other, front=front))

    def tensor(self, other: Chi) -> Chi:
        if not isinstance(other, Chi):
            other = Chi(other)
        return self._tensor(self, other)

    def expand(self, other: Chi) -> Chi:
        if not isinstance(other, Chi):
            other = Chi(other)
        return self._tensor(other, self)

    @classmethod
    def _tensor(cls, a, b):
        ret = _copy.copy(a)
        ret._op_shape = a._op_shape.tensor(b._op_shape)
        ret._data = np.kron(a._data, b.data)
        return ret


# Update docstrings for API docs
generate_apidocs(Chi)
